/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>

GstElementDetails gst_bin_details = { 
  "Generic bin",
  "Bin",
  "Simple container object",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


static gboolean gst_bin_change_state(GstElement *element,
                                     GstElementState state);
static gboolean gst_bin_change_state_type(GstBin *bin,
                                          GstElementState state,
                                          GtkType type);

/* Bin signals and args */
enum {
  OBJECT_ADDED,
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_bin_class_init(GstBinClass *klass);
static void gst_bin_init(GstBin *bin);


static GstElementClass *parent_class = NULL;
static guint gst_bin_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_bin_get_type(void) {
  static GtkType bin_type = 0;

  if (!bin_type) {
    static const GtkTypeInfo bin_info = {
      "GstBin",
      sizeof(GstBin),
      sizeof(GstBinClass),
      (GtkClassInitFunc)gst_bin_class_init,
      (GtkObjectInitFunc)gst_bin_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    bin_type = gtk_type_unique(GST_TYPE_ELEMENT,&bin_info);
  }
  return bin_type;
}

static void
gst_bin_class_init(GstBinClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gst_bin_signals[OBJECT_ADDED] =
    gtk_signal_new("object_added",GTK_RUN_FIRST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstBinClass,object_added),
                   gtk_marshal_NONE__POINTER,GTK_TYPE_NONE,1,
                   GTK_TYPE_POINTER);
  gtk_object_class_add_signals(gtkobject_class,gst_bin_signals,LAST_SIGNAL);

  klass->change_state_type = gst_bin_change_state_type;

  gstelement_class->change_state = gst_bin_change_state;
}

static void gst_bin_init(GstBin *bin) {
  bin->numchildren = 0;
  bin->children = NULL;
}

/**
 * gst_bin_new:
 * @name: name of new bin
 *
 * Create a new bin with given name.
 *
 * Returns: new bin
 */
GstElement *gst_bin_new(gchar *name) {
  GstElement *bin = GST_ELEMENT(gtk_type_new(GST_TYPE_BIN));
  gst_object_set_name(GST_OBJECT(bin),name);
  return bin;
}

/**
 * gst_bin_add:
 * @bin: #GstBin to add element to
 * @element: #GstElement to add to bin
 *
 * Add the given element to the bin.  Set the elements parent, and thus
 * add a reference.
 */
void gst_bin_add(GstBin *bin,GstElement *element) {
  g_return_if_fail(bin != NULL);
  g_return_if_fail(GST_IS_BIN(bin));
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));

  bin->children = g_list_append(bin->children,element);
  bin->numchildren++;
  gst_object_set_parent(GST_OBJECT(element),GST_OBJECT(bin));

  gtk_signal_emit(GTK_OBJECT(bin),gst_bin_signals[OBJECT_ADDED],element);
}

/**
 * gst_bin_remove:
 * @bin: #Gstbin to remove element from
 * @element: #GstElement to remove
 *
 * Remove the element from its associated bin, unparenting as well.
 */
void gst_bin_remove(GstBin *bin,GstElement *element) {
  g_return_if_fail(bin != NULL);
  g_return_if_fail(GST_IS_BIN(bin));
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));
  g_return_if_fail(bin->children != NULL);

  gst_object_unparent(GST_OBJECT(element));
  bin->children = g_list_remove(bin->children,element);
  bin->numchildren--;
}


static gboolean gst_bin_change_state(GstElement *element,
                                     GstElementState state) {
  GstBin *bin;
  GList *children;
  GstElement *child;

//  g_print("gst_bin_change_state(\"%s\",%d);\n",
//          gst_object_get_name(GST_OBJECT(bin)),state);

  g_return_if_fail(GST_IS_BIN(element));
  bin = GST_BIN(element);
  g_return_if_fail(bin->numchildren != 0);

//  g_print("-->\n");
  children = bin->children;
  while (children) {
    child = GST_ELEMENT(children->data);
//    g_print("gst_bin_change_state setting state on \"%s\"\n",
//            gst_object_get_name(GST_OBJECT(child)));
    if (!gst_element_set_state(child,state)) {
      g_print("child %p failed to set state 0x%08x\n",child,state);
      return FALSE;
    }
//    g_print("\n");
    children = g_list_next(children);
  }
//  g_print("<-- \"%s\"\n",gst_object_get_name(GST_OBJECT(bin)));

  if (GST_ELEMENT_CLASS(parent_class)->change_state)
    return GST_ELEMENT_CLASS(parent_class)->change_state(element,state);
  return TRUE;
}

static gboolean gst_bin_change_state_type(GstBin *bin,
                                          GstElementState state,
                                          GtkType type) {
  GList *children;
  GstElement *child;

//  g_print("gst_bin_change_state_type(\"%s\",%d,%d);\n",
//          gst_object_get_name(GST_OBJECT(bin)),state,type);

  g_return_if_fail(GST_IS_BIN(bin));
  g_return_if_fail(bin->numchildren != 0);

//  g_print("-->\n");
  children = bin->children;
  while (children) {
    child = GST_ELEMENT(children->data);
    if (GST_IS_BIN(child)) {
      if (!gst_bin_set_state_type(GST_BIN(child),state,type))
        return FALSE;
    } else if (GTK_CHECK_TYPE(child,type)) {
      if (!gst_element_set_state(child,state))
        return FALSE;
    }
//    g_print("\n");
    children = g_list_next(children);
  }
//  g_print("<-- \"%s\"\n",gst_object_get_name(GST_OBJECT(bin)));
  if (type == GST_TYPE_BIN)
    gst_element_change_state(GST_ELEMENT(bin),state);

  return TRUE;
}


gboolean gst_bin_set_state_type(GstBin *bin,
                                GstElementState state,
                                GtkType type) {
  GstBinClass *oclass;

//  g_print("gst_bin_set_state_type(\"%s\",%d,%d)\n",
//          gst_object_get_name(GST_OBJECT(bin)),state,type);

  g_return_if_fail(bin != NULL);
  g_return_if_fail(GST_IS_BIN(bin));

  oclass = GST_BIN_CLASS(GTK_OBJECT(bin)->klass);

  if (oclass->change_state_type);
    (oclass->change_state_type)(bin,state,type);
}
